from django.db import models
import uuid
import json

from django_celery_beat.models import CrontabSchedule, PeriodicTask
import datetime
import pytz
from . import TASK_WEIGHT_SAMPLE_SIZE
from django.conf import settings


class CeleryQueue(models.Model):
    """
    A way to control which queue goes where.
    The default queue django-celery uses is "celery".
    If you want to follow a specific queue, call celery with
    the "-Q queue_name" parameter.
    """
    queue_name = models.CharField(max_length=4096)

    def __str__(self):
        return self.queue_name


class DefaultBaseModel(models.Model):
    """
    Abstract base model that is used to add `created_date`
    and `modified_date` fields to all decendent models.
    Also uses UUID4 for the ID
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created_date = models.DateTimeField(auto_now_add=True)
    modified_date = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class Cog(DefaultBaseModel):
    """
    A way to keep track of registered cogs
    """
    name = models.CharField(max_length=4096)

    def __str__(self):
        return self.name


class CogErrorHandler(DefaultBaseModel):
    """
    A way to keep track of registered cogs error handlers
    """
    name = models.CharField(max_length=4096)

    def __str__(self):
        return self.name


class Pipeline(DefaultBaseModel):
    """
    A collection of stages all linked together.
    A pipeline is what gets called via a cron schedule
    """
    schedule = models.ForeignKey(CrontabSchedule, null=True, blank=True, on_delete=models.CASCADE)
    name = models.CharField(max_length=4096)
    enabled = models.BooleanField(default=True)
    prevent_overlapping_runs = models.BooleanField(default=True)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        """
        Upon saving, make sure we create the cron schedule.
        """
        # then create the PeriodicTask
        #   NOTE: Use the UUID of this pipeline as the name,
        #   which will make deleting it safer
        if self.schedule:
            periodic_task, created = PeriodicTask.objects.get_or_create(
                name=self.id,
                defaults={
                    'crontab': self.schedule
                }
            )
            periodic_task.crontab = self.schedule
            periodic_task.task = 'django_cog.launch_pipeline'
            periodic_task.args = json.dumps([str(self.id)])
            periodic_task.kwargs = json.dumps({
                'pipeline_id': str(self.id)
            })
            periodic_task.enabled = self.enabled
            periodic_task.save()
        else:
            # remove any periodic tasks that may have been previously
            # created
            PeriodicTask.objects.filter(
                name=self.id
            ).delete()

        super(Pipeline, self).save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """
        Upon deleting, make sure to remove the created PeriodicTask.
        """
        # delete the periodic task that was created from this pipeline
        if PeriodicTask.objects.filter(name=self.id).exists():
            PeriodicTask.objects.filter(name=self.id).delete()
        
        # then delete the pipeline object
        super(Pipeline, self).delete(*args, **kwargs)


class Stage(DefaultBaseModel):
    """
    A stage is an ordered element of a pipeline, and will
    consist of a collection of cogs.  Cogs in the same
    stage are processes that can be ran independently of each other.

    When one stage of a pipeline completes, the next stage is began.

    Here, "proceeds" will point to the PREVIOUS stage.
        - NULL will mean it is the first stage to launch at the start of the pipeline.
    """
    name = models.CharField(max_length=4096)
    pipeline = models.ForeignKey(Pipeline, related_name='stages', on_delete=models.CASCADE)
    launch_after_stage = models.ManyToManyField('Stage', related_name='next_stage', blank=True)

    def __str__(self):
        return f"{self.pipeline} - {self.name}"
    
    class Meta:
        ordering = ['pipeline', 'name']


class Task(DefaultBaseModel):
    """
    A specific registered function to tie to a stage.

    - Prevent Overlapping Calls:
        If true, the function defined in the Cog will not be called if we detect
        a TaskRun has not yet completed with the same Cog definition.
    """
    name = models.CharField(max_length=4096, blank=True, null=True)
    cog = models.ForeignKey(Cog, related_name='assigned_tasks', on_delete=models.CASCADE)
    stage = models.ForeignKey(Stage, related_name='assigned_tasks', on_delete=models.CASCADE)
    error_handler = models.ForeignKey(CogErrorHandler, related_name='tasks', on_delete=models.SET_NULL, null=True, blank=True)
    arguments_as_json = models.TextField(blank=True, null=True)
    prevent_overlapping_calls = models.BooleanField(default=True)
    enabled = models.BooleanField(default=True)
    queue = models.ForeignKey(CeleryQueue, related_name='assigned_tasks', on_delete=models.SET_NULL, default=1, null=True)
    weight = models.DecimalField(decimal_places=5, max_digits=12, default=1.0)
    critical = models.BooleanField(default=True)

    def __str__(self):
        if self.name:
            return self.name
        return str(self.cog)


# //-- Below are records kept for execution runs --//
class EntityRun(DefaultBaseModel):
    """
    Base model for runs.  Includes a start and completed on.
    """
    STATUS_OPTIONS = [
        ('Queued', 'Queued'),
        ('Running', 'Running'),
        ('Failed', 'Failed'),
        ('Completed', 'Completed'),
        ('Canceled', 'Canceled')
    ]
    started_on = models.DateTimeField(auto_now_add=True)
    completed_on = models.DateTimeField(null=True, blank=True)
    status = models.CharField(max_length=10, choices=STATUS_OPTIONS, default='Queued', null=True)
    
    def __str_runtimes__(self):
        string = f"[{self.started_on}]"
        if self.completed_on:
            string += ' COMPLETED'
        return string

    def runtime(self):
        if self.completed_on:
            return (self.completed_on - self.started_on)
        return None
    
    class Meta:
        ordering = ['-completed_on', '-started_on']

class PipelineRun(EntityRun):
    """
    Specific execution run of a pipeline.
    """
    pipeline = models.ForeignKey(Pipeline, related_name='runs', on_delete=models.CASCADE)
    success = models.NullBooleanField()

    def __str__(self):
        return f"{str(self.pipeline)} {self.__str_runtimes__()}"
    
    def save(self, *args, **kwargs):
        # how many tasks need to be completed?
        required = self.pipeline.stages.all().count()
        completed = self.stage_runs.filter(
            completed_on__isnull=False
        ).count()
        if required == completed:
            self.completed_on = datetime.datetime.now(tz=pytz.UTC)

            # did we complete successfully?  if we're here and haven't been set to false yet,
            # then we can assume it's all good
            if self.success is None:
                self.success = True

            # if we were running (haven't failed critically), update to completed
            if self.status == 'Running':
                self.status = 'Completed'

            # do we want to auto-update weight?
            if not hasattr(settings, 'DJANGO_COG_AUTO_WEIGHT') or settings.DJANGO_COG_AUTO_WEIGHT:
                # pull the weight sample size
                sample_size = TASK_WEIGHT_SAMPLE_SIZE
                if hasattr(settings, 'DJANGO_COG_TASK_WEIGHT_SAMPLE_SIZE'):
                    sample_size = settings.DJANGO_COG_TASK_WEIGHT_SAMPLE_SIZE

                # make a way to compute runtime from SQL:
                duration = models.ExpressionWrapper(
                    models.F('completed_on') - models.F('started_on'),
                    output_field=models.fields.DurationField()
                )

                # if we're completed, let's update the weights of all our tasks
                for stage in self.pipeline.stages.all():
                    # update each tasks in the stage
                    for task in stage.assigned_tasks.all():
                        # get a sample of this tasks runs
                        average_weight = task.runs.all()[:sample_size].annotate(
                            duration=duration
                        ).aggregate(models.Avg('duration'))['duration__avg'].total_seconds()
                        task.weight  = average_weight
                        task.save()
        super(PipelineRun, self).save(*args, **kwargs)

class StageRun(EntityRun):
    """
    Specific execution run of a stage.
    """
    pipeline_run = models.ForeignKey(PipelineRun, related_name='stage_runs', on_delete=models.CASCADE)
    stage = models.ForeignKey(Stage, related_name='runs', on_delete=models.CASCADE)
    success = models.NullBooleanField()

    def __str__(self):
        return f"{str(self.stage)} {self.__str_runtimes__()}"
    
    def save(self, *args, **kwargs):
        # how many tasks need to be completed?
        required = self.stage.assigned_tasks.filter(enabled=True).count()
        completed = self.task_runs.filter(
            completed_on__isnull=False
        ).count()
        if required == completed and self.completed_on is None:
            self.completed_on = datetime.datetime.now(tz=pytz.UTC)

            # did we complete it successfully?  default is yes unless we've been set
            # to false by a failed task already
            if self.success is None:
                self.success = True
            
            # if we were running (haven't failed critically), update to completed
            if self.status == 'Running':
                self.status = 'Completed'

            # call the super to make sure that when we launch the next stage,
            # it knows we're done, and will be counted then
            super(StageRun, self).save(*args, **kwargs)

            # make sure that any crashed tasks weren't critical...
            required_completed = True
            for task in self.task_runs.filter(success=False):
                if task.task.critical:
                    print("!! Found crashed tasks that are marked critical, so we are not launching the next stage!")
                    required_completed = False
            
            # and make sure we haven't had a call for cancellation yet...
            canceled = False
            if self.pipeline_run.status == 'Canceled':
                print("!! Found pipeline run in canceled state, so we are not launching the next stage.")
                self.status = 'Canceled'
                canceled = True

            # if this is the case, we can call the next stages!
            if required_completed and not canceled and self.stage.next_stage.all().count() > 0:
                from django_cog import launch_stage
                for stage in self.stage.next_stage.all():
                    launch_stage.delay(stage_id=stage.id, pipeline_run_id=self.pipeline_run.id)

        else:
            # if we were canceled, the completed on time would be filled in.  we just need to
            # mark ourselves as canceled now as well:
            if self.pipeline_run.status == 'Canceled':
                self.status = 'Canceled'

            # make sure to save this run in case we haven't completed yet...so our tasks
            # we're about to launch know who they belong to
            super(StageRun, self).save(*args, **kwargs)

        # trigger the pipeline run save just in case this is the last stage
        # note: we must do this after calling the super, so the pipeline run can pick up
        #   that we've completed this stage run
        self.pipeline_run.save()


class TaskRun(EntityRun):
    """
    Specific execution run of a task.
    """
    stage_run = models.ForeignKey(StageRun, related_name='task_runs', on_delete=models.CASCADE)
    task = models.ForeignKey(Task, related_name='runs', on_delete=models.CASCADE)
    success = models.NullBooleanField()

    def __str__(self):
        return f"{str(self.task)} {self.__str_runtimes__()}"


class CogError(DefaultBaseModel):
    """
    Support for default cog error handler.
    Stores information about the error in the database.
    """
    task_run = models.ForeignKey(TaskRun, related_name='errors', on_delete=models.CASCADE, null=True, blank=True)
    timestamp = models.DateTimeField(auto_created=True, auto_now_add=True)
    error_type = models.CharField(max_length=1024, blank=True, null=True)
    traceback = models.TextField(blank=True, null=True)

    def __str__(self):
        return f"{self.task_run.task.name} - {self.timestamp.strftime('%Y-%m-%d %H:%M:%S')}"
    class Meta:
        ordering = ['-timestamp']


from django.contrib import admin
from django.shortcuts import HttpResponseRedirect
from . import launch_pipeline
from nested_inline.admin import NestedStackedInline, NestedModelAdmin
from .models import (
    Cog,
    CogErrorHandler,
    Pipeline,
    Stage,
    Task,
    PipelineRun,
    StageRun,
    TaskRun,
    CeleryQueue,
    CogError
)

# Register your models here.
@admin.register(CeleryQueue)
class CeleryQueueAdmin(admin.ModelAdmin):
    list_display = ["queue_name"]
    search_fields = ["queue_name"]

@admin.register(Cog)
class CogAdmin(admin.ModelAdmin):
    list_display = ["name"]
    search_fields = ["name"]

@admin.register(CogError)
class CogErrorAdmin(admin.ModelAdmin):
    list_display = ["get_task_name", "error_type", "timestamp"]

    def get_task_name(self, obj):
        return obj.task_run.task.name

@admin.register(CogErrorHandler)
class CogErrorHandlerAdmin(admin.ModelAdmin):
    list_display = ["name"]
    search_fields = ["name"]

class TasksInline(NestedStackedInline):
    model = Task
    extra = 0
class StagesInLine(NestedStackedInline):
    model = Stage
    inlines = [TasksInline,]
    extra = 0

@admin.register(Pipeline)
class PipelineAdmin(NestedModelAdmin):
    list_display = ["name", "schedule", "enabled"]
    search_fields = ["name"]
    inlines = [StagesInLine,]
    change_form_template = 'change_form.html'

    def response_change(self, request, obj):
        """
        Override to support the "Launch Now" button.
        """
        if "_launchbutton" in request.POST:
            launch_pipeline.apply_async(
                queue='celery',
                kwargs = {
                    'pipeline_id': obj.id, 
                    'user_initiated': True
                })
            self.message_user(request, "Pipeline launch has been initiated.")
            return HttpResponseRedirect(".")
        return super().response_change(request, obj)


@admin.register(Stage)
class StageAdmin(admin.ModelAdmin):
    list_display = ["name", "pipeline"]
    search_fields = ["name"]
    inlines = [TasksInline,]


@admin.register(Task)
class TaskAdmin(admin.ModelAdmin):
    list_display = ["task_name", "cog", "stage", "pipeline", "queue", "weight"]
    search_fields = ["cog", "queue"]

    def pipeline(self, obj):
        return obj.stage.pipeline
    
    def task_name(self, obj):
        if obj.name:
            return obj.name
        return obj.cog


class StageRunInLine(admin.TabularInline):
    model = StageRun
    extra = 0
    fk_name = 'pipeline_run'
    fieldsets = (
        (
            None, {
                'fields': ('stage', 'started_on', 'completed_on', 'runtime')
            }
        ),(
            'None', {'fields': ()}
        )
    )
    readonly_fields = ('stage', 'started_on', 'completed_on', 'runtime')
    # this is merely for record keeping, and so we shouldn't
    # be adding any through the admin
    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

class TaskRunInLine(admin.TabularInline):
    model = TaskRun
    extra = 0
    fk_name = 'stage_run'
    fieldsets = (
        (
            None, {
                'fields': ('task', 'started_on', 'completed_on', 'runtime')
            }
        ),(
            'None', {'fields': ()}
        )
    )
    readonly_fields = ('task', 'started_on', 'completed_on', 'runtime')
    # this is merely for record keeping, and so we shouldn't
    # be adding any through the admin
    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

@admin.register(PipelineRun)
class PipelineRunAdmin(admin.ModelAdmin):
    list_display = ["pipeline", "status", "started_on", "completed_on", "runtime"]
    search_fields = ["pipeline", "started_on"]
    inlines = [StageRunInLine,]
    change_form_template = 'cancel_run.html'

    def response_change(self, request, obj):
        """
        Override to support the "Cancel Run" button.
        """
        if "_cancelbutton" in request.POST:
            obj.status = 'Canceled'
            obj.save()
            self.message_user(request, "Pipeline run will halt after its currently running tasks complete.")
            return HttpResponseRedirect(".")
        return super().response_change(request, obj)

@admin.register(StageRun)
class StageRunAdmin(admin.ModelAdmin):
    list_display = ["stage", "pipeline", "status", "started_on", "completed_on", "runtime"]
    search_fields = ["stage", "pipeline", "started_on"]
    inlines = [TaskRunInLine,]

    def pipeline(self, obj):
        return obj.pipeline_run.pipeline


@admin.register(TaskRun)
class TaskRunAdmin(admin.ModelAdmin):
    list_display = ["task", "stage", "pipeline", "status", "started_on", "completed_on", "runtime"]
    search_fields = ["task", "stage", "pipeline", "started_on"]

    def pipeline(self, obj):
        return obj.stage_run.pipeline_run.pipeline
    
    def stage(self, obj):
        return obj.stage_run.stage



# Generated by Django 2.2.23 on 2021-09-15 14:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0014_auto_20210915_1035'),
    ]

    operations = [
        migrations.AddField(
            model_name='cogerror',
            name='error_type',
            field=models.CharField(blank=True, max_length=1024, null=True),
        ),
    ]



# Generated by Django 2.2.23 on 2021-09-15 13:27

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0012_cogerror'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='error_handler',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tasks', to='django_cog.CogErrorHandler'),
        ),
    ]



# Generated by Django 2.2.20 on 2021-04-26 19:50

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('django_celery_beat', '0015_edit_solarschedule_events_choices'),
    ]

    operations = [
        migrations.CreateModel(
            name='Cog',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=4096)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='EntityRun',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True)),
                ('started_on', models.DateTimeField(auto_now_add=True)),
                ('completed_on', models.DateTimeField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Pipeline',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=4096)),
                ('enabled', models.BooleanField(default=True)),
                ('prevent_overlapping_runs', models.BooleanField(default=True)),
                ('schedule', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_celery_beat.CrontabSchedule')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Stage',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=4096)),
                ('launch_after_stage', models.ManyToManyField(related_name='next_stage', to='django_cog.Stage', blank=True)),
                ('pipeline', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='stages', to='django_cog.Pipeline')),
            ],
            options={
                'ordering': ['pipeline', 'name'],
            },
        ),
        migrations.CreateModel(
            name='PipelineRun',
            fields=[
                ('entityrun_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='django_cog.EntityRun')),
                ('pipeline', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='runs', to='django_cog.Pipeline')),
            ],
            options={
                'abstract': False,
            },
            bases=('django_cog.entityrun',),
        ),
        migrations.CreateModel(
            name='StageRun',
            fields=[
                ('entityrun_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='django_cog.EntityRun')),
                ('pipeline_run', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='stage_runs', to='django_cog.PipelineRun')),
                ('stage', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='runs', to='django_cog.Stage')),
            ],
            options={
                'abstract': False,
            },
            bases=('django_cog.entityrun',),
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(blank=True, max_length=4096, null=True)),
                ('arguments_as_json', models.TextField(blank=True, null=True)),
                ('prevent_overlapping_calls', models.BooleanField(default=True)),
                ('enabled', models.BooleanField(default=True)),
                ('cog', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='assigned_tasks', to='django_cog.Cog')),
                ('stage', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='assigned_tasks', to='django_cog.Stage')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TaskRun',
            fields=[
                ('entityrun_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='django_cog.EntityRun')),
                ('stage_run', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='task_runs', to='django_cog.StageRun')),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='runs', to='django_cog.Task')),
            ],
            options={
                'abstract': False,
            },
            bases=('django_cog.entityrun',),
        ),
    ]


# Generated by Django 2.2.23 on 2021-05-19 13:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0003_task_queue'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='entityrun',
            options={'ordering': ['-completed_on', '-started_on']},
        ),
        migrations.AddField(
            model_name='task',
            name='weight',
            field=models.DecimalField(decimal_places=5, default=1.0, max_digits=10),
        ),
    ]



from .celery import celery_app as celery_app
import datetime
import pytz
import json
import traceback
import inspect

__all__ = ['celery_app']

# default sample for calculating weight of tasks
TASK_WEIGHT_SAMPLE_SIZE = 10


def CogRegistration():
    """
    Decorator used to register tasks.

    Shamelessly stolen from: https://stackoverflow.com/questions/5707589/calling-functions-by-array-index-in-python/5707605#5707605
    """
    registry = {}
    def registrar(func):
        registry[func.__name__] = func
        return func  # normally a decorator returns a wrapped function, 
                     # but here we return func unmodified, after registering it
    registrar.all = registry
    return registrar

cog = CogRegistration()


def CogErrorHandlerRegistration():
    """
    Decorator used to register error handlers.

    Shamelessly stolen from: https://stackoverflow.com/questions/5707589/calling-functions-by-array-index-in-python/5707605#5707605
    """
    registry = {}
    def registrar(func):
        registry[func.__name__] = func
        return func  # normally a decorator returns a wrapped function, 
                     # but here we return func unmodified, after registering it
    registrar.all = registry
    return registrar

cog_error_handler = CogErrorHandlerRegistration()


@cog_error_handler
def defaultCogErrorHandler(error, task_run=None):
    """
    Default cog error handler.  Records the error's type, time, and traceback
    into the database.
    """
    from .models import CogError
    error_info = "".join(traceback.format_exception(etype=None, value=error, tb=error.__traceback__))
    CogError.objects.create(
        task_run=task_run,
        traceback=error_info,
        error_type=type(error).__name__
    )


# //-- Below are the Celery tasks used to launch Tasks, Stages, and Pipelines --//
@celery_app.task
def launch_task(task_id, stage_run_id):
    """
    Main celery task called when launching a new stage.
    """
    from .models import Task, TaskRun, StageRun
    # look up the task
    try:
        task = Task.objects.get(id=task_id)
    except Task.DoesNotExist:
        print("!! Could not find task called.  Aboring.")
        exit()
    
    # is this enabled?
    if not task.enabled:
        return
    
    # look up the stage ruin
    try:
        stage_run = StageRun.objects.get(id=stage_run_id)
    except StageRun.DoesNotExist:
        print("!! Could not find stage run called.  Aboring.")
        exit()
    
    # make sure a previous task hasn't failed, if it's critical
    for previous_task in stage_run.task_runs.filter(success=False):
        if previous_task.task.critical:
            print("!! A previously crashed task was critical to the pipeline, so we're aborting.")
            exit()
    
    # make sure we haven't been canceled yet
    if stage_run.pipeline_run.status == 'Canceled':
        print("!! The pipeline run has been canceled.  Aborting task.")
        exit()
    
    # do we already have a task running that hasn't completed (and do we care)?
    # note: this check is cross-stage checking.  Really, this shouldn't ever fire off
    # unless we start a new run of the Pipeline before a previous one has finished,
    # OR this task is part of a different pipeline.
    if task.prevent_overlapping_calls and TaskRun.objects.filter(task=task, completed_on__isnull=True).exists():
        print("!! Another instance of this task has not completed yet.  Aborting.")
        exit()
    
    # otherwise, create a new pipeline run
    task_run = TaskRun.objects.create(
        stage_run=stage_run,
        task=task
    )

    # get the function in our `cog` that matches the name
    try:
        kwargs = {}
        if task.arguments_as_json:
            kwargs = json.loads(task.arguments_as_json)
        task.status = 'Running'
        task.save()
        cog.all[task.cog.name](**kwargs)
    except Exception as e:
        # if there's an error handler for this function, pass what ever sort of Exception we got
        # into the handler
        if task.error_handler:
            try:
                signature = inspect.signature(cog_error_handler.all[task.error_handler.name])
                if 'task_run' in signature.parameters:
                    cog_error_handler.all[task.error_handler.name](e, task_run=task_run)
                else:
                    cog_error_handler.all[task.error_handler.name](e)
            except Exception as error_handler_error:
                print("!! Failed to execute error handler!  Make sure it has only one parameter to take in (the error Exception)")
                print(error_handler_error)
        else:
            # do we at least have the default one that should come with this library? (like..it's literally defined above)
            if 'defaultCogErrorHandler' in cog_error_handler.all:
                # then we can use this as our default
                cog_error_handler.all['defaultCogErrorHandler'](e, task_run=task_run)

        # save this task run as having failed
        task_run.success = False
        task_run.status = 'Failed'
        task_run.save()

        # also mark the stage and pipelines
        stage_run.success = False

        # only mark the stage's status as failed if this task is critical
        if task.critical:
            stage_run.status = 'Failed'
            stage_run.pipeline_run.status = 'Failed'

        stage_run.save()
        stage_run.pipeline_run.success = False
        stage_run.pipeline_run.save()
    
    # mark that this task is completed
    task_run.completed_on = datetime.datetime.now(tz=pytz.UTC)

    # mark it successful unless it failed above
    if task_run.success is None:
        task_run.status = 'Completed'
        task_run.success = True
    task_run.save()

    # we need to check if the stage and pipeline has failed, and if we're the last running.
    # this way we can say "this is when we completed"
    # be sure to refresh from the DB in case another task has made changes:
    stage_run.refresh_from_db()
    if (stage_run.status == 'Failed' or stage_run.pipeline_run.status == 'Canceled') and stage_run.task_runs.all().first() == task_run:
        # then we're the last one, so set the completed time for the stage run and pipeline run
        stage_run.completed_on = datetime.datetime.now(tz=pytz.UTC)
        stage_run.pipeline_run.completed_on = datetime.datetime.now(tz=pytz.UTC)
        stage_run.pipeline_run.save()

    # trigger the save on our stage run so we can check if we're done with all of our tasks
    stage_run.save()

@celery_app.task
def launch_stage(stage_id, pipeline_run_id):
    """
    Main celery task called when launching a new stage.
    """
    from .models import Stage, StageRun, PipelineRun
    # look up the stage
    try:
        stage = Stage.objects.get(id=stage_id)
    except Stage.DoesNotExist:
        print("!! Could not find stage called.  Aboring.")
        exit()
    
    # look up the pipeline run
    try:
        pipeline_run = PipelineRun.objects.get(id=pipeline_run_id)
    except PipelineRun.DoesNotExist:
        print("!! Could not find pipeline run called.  Aboring.")
        exit()
    
    # make sure all previous stage runs have completed
    for required_stage in stage.launch_after_stage.all():
        if not StageRun.objects.filter(
            stage=required_stage,
            pipeline_run=pipeline_run,
            completed_on__isnull=False
        ).exists():
            # if there isn't a run that is completed, then abort
            # (once this required stage run completes, this function
            #   will be called again)
            return
    
    # otherwise, create a new pipeline run
    stage_run = StageRun.objects.create(
        stage=stage,
        pipeline_run=pipeline_run
    )
    stage_run.status = 'Running'
    stage_run.save()

    # order tasks by their weight
    # this ensures the longest tasks get called first, which helps improve overall runtime
    tasks = stage.assigned_tasks.filter(enabled=True).order_by('-weight')
    # launch all tasks concurrently
    for task in tasks:
        # assume default ("celery") queue, unless specified
        queue = 'celery'
        if task.queue:
            queue = task.queue.queue_name
        launch_task.apply_async(
            queue = queue,
            kwargs = {
                'task_id': task.id,
                'stage_run_id': stage_run.id
            }
        )

@celery_app.task
def launch_pipeline(*args, **kwargs):
    """
    Main django-celery-beat function called when a pipeline is ready to launch
    """
    if 'pipeline_id' not in kwargs:
        print("!! Incorrect calling of function.  Could not find pipeline identifier.")
        exit()

    from django.conf import settings
    from .models import Pipeline, PipelineRun
    # look up the pipeline
    try:
        pipeline = Pipeline.objects.get(id=kwargs['pipeline_id'])
    except Pipeline.DoesNotExist:
        print("!! Could not find pipeline called.  Aboring.")
        exit()
    
    # is this pipeline enabled, and do we check for it?
    if 'user_initiated' not in kwargs and not pipeline.enabled:
        print("Pipeline not enabled.  Aborting.")
        return
    
    # is there another run of this pipline still going? (and do we care?)
    if pipeline.prevent_overlapping_runs and PipelineRun.objects.filter(
        pipeline=pipeline,
        status='Running'
    ).exists():
        print("!! Another instance of this pipeline has not completed.  Aborting.")
        return
    
    # also check if there are failed runs, and if we should care about that
    if pipeline.prevent_overlapping_runs and PipelineRun.objects.filter(
        pipeline=pipeline
    ).exists() and PipelineRun.objects.filter(
        pipeline=pipeline
    ).first().status == 'Failed' \
    and (not hasattr(settings, 'DJANGO_COG_OVERLAP_FAILED') or settings.DJANGO_COG_OVERLAP_FAILED == False):
        print("!! The last pipeline run failed and we cannot overlap failed runs.  Remove the failed pipeline run or set `DJANGO_COG_OVERLAP_FAILED` to True in your Django settings.  Aborting.")
        return

    # otherwise, create a new pipeline run
    pipeline_run = PipelineRun.objects.create(
        pipeline=pipeline
    )
    pipeline_run.status = 'Running'
    pipeline_run.save()

    # start the first stage
    first_stages = pipeline.stages.filter(
        launch_after_stage=None
    )
    for stage in first_stages:
        # theoritically, you should never have more than one stage,
        # because all tasks could be bundled into a single stage
        launch_stage.delay(stage_id=stage.id, pipeline_run_id=pipeline_run.id)



# Generated by Django 2.2.23 on 2021-05-19 12:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0002_auto_20210519_0824'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='queue',
            field=models.ForeignKey(default=1, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='assigned_tasks', to='django_cog.CeleryQueue'),
        ),
    ]



# Generated by Django 2.2.23 on 2021-09-15 19:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0017_auto_20210915_1533'),
    ]

    operations = [
        migrations.AlterField(
            model_name='entityrun',
            name='status',
            field=models.CharField(choices=[('Queued', 'Queued'), ('Running', 'Running'), ('Failed', 'Failed'), ('Completed', 'Completed'), ('Canceled', 'Canceled')], default='Queued', max_length=10, null=True),
        ),
    ]



# Generated by Django 2.2.23 on 2021-05-25 12:09

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0005_auto_20210519_1006'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='taskrun',
            options={},
        ),
        migrations.AlterField(
            model_name='pipeline',
            name='schedule',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='django_celery_beat.CrontabSchedule'),
        ),
    ]



# Generated by Django 2.2.23 on 2021-09-14 18:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0007_auto_20210525_0810'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='critical',
            field=models.BooleanField(default=True),
        ),
    ]



# Generated by Django 2.2.23 on 2021-05-25 12:10

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0006_auto_20210525_0809'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pipeline',
            name='schedule',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='django_celery_beat.CrontabSchedule'),
        ),
    ]



# Generated by Django 2.2.23 on 2021-05-19 12:24

from django.db import migrations, models


def create_celery_queue(apps, schema_editor):
    """
    Creates a default "celery" queue.  This is what the next
    migration will use as the default.
    """
    CeleryQueue = apps.get_model('django_cog', 'CeleryQueue')
    CeleryQueue.objects.create(
        id=1,
        queue_name='celery'
    )

class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='CeleryQueue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('queue_name', models.CharField(max_length=4096)),
            ],
        ),
        migrations.AlterModelOptions(
            name='taskrun',
            options={'ordering': ['completed_on']},
        ),
        migrations.RunPython(create_celery_queue)
    ]



# Generated by Django 2.2.23 on 2021-05-19 14:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0004_auto_20210519_0931'),
    ]

    operations = [
        migrations.AlterField(
            model_name='task',
            name='weight',
            field=models.DecimalField(decimal_places=5, default=1.0, max_digits=12),
        ),
    ]



# Generated by Django 2.2.23 on 2021-09-15 13:17

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0010_auto_20210914_1500'),
    ]

    operations = [
        migrations.CreateModel(
            name='CogErrorHandler',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=4096)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="django-cog",
    version="1.3.2",
    description="Django library for launching pipelines of multiple stages and parallel tasks.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/david-pettifor-nd/django_cog.git",
    author="David W Pettifor",
    author_email="dpettifo@nd.edu",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Framework :: Django :: 2.2"
    ],
    packages=["django_cog"],
    include_package_data=True,
    install_requires=["celery==5.2.0", "django-celery-beat", "django-nested-inline>=0.4.2"]
)



# Generated by Django 2.2.23 on 2021-09-15 14:35

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0013_task_error_handler'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cogerror',
            name='task_run',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='errors', to='django_cog.TaskRun'),
        ),
    ]



# Generated by Django 2.2.23 on 2021-09-14 19:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0009_taskrun_success'),
    ]

    operations = [
        migrations.AddField(
            model_name='pipelinerun',
            name='success',
            field=models.NullBooleanField(),
        ),
        migrations.AddField(
            model_name='stagerun',
            name='success',
            field=models.NullBooleanField(),
        ),
    ]



from django.apps import AppConfig


def create_cog_records():
    """
    This function loops through any known
    registered cogs and creates Cog objects
    for them.
    """
    try:
        from .models import Cog
        from . import cog
        active_cogs = []
        for cog_name, function in cog.all.items():
            try:
                Cog.objects.get_or_create(
                    name=cog_name
                )
                active_cogs.append(cog_name)
            except Cog.MultipleObjectsReturned:
                print(f"ERROR: Found multiple functions with the same name `{cog_name}`.  Change these function names to be unique in order to register them.")

    except Exception as e:
        print(e)
        print("Failed to clean up cogs.  Maybe try running migrations?")


def create_cog_error_handler_records():
    """
    This function loops through any known
    registered cog error handlers and creates 
    CogErrorHandler objects for them.
    """
    try:
        from .models import CogErrorHandler
        from . import cog_error_handler
        active_error_handlers = []
        for handler_name, function in cog_error_handler.all.items():
            try:
                CogErrorHandler.objects.get_or_create(
                    name=handler_name
                )
                active_error_handlers.append(handler_name)
            except CogErrorHandler.MultipleObjectsReturned:
                print(f"ERROR: Found multiple functions with the same name `{handler_name}`.  Change these function names to be unique in order to register them.")

    except Exception as e:
        print(e)
        print("Failed to clean up cog error handlers.  Maybe try running migrations?")


class DjangoCogConfig(AppConfig):
    name = 'django_cog'

    def ready(self):
        create_cog_records()
        create_cog_error_handler_records()



# Generated by Django 2.2.23 on 2021-09-15 14:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0015_cogerror_error_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cogerror',
            name='timestamp',
            field=models.DateTimeField(auto_created=True, auto_now_add=True),
        ),
    ]



# Generated by Django 2.2.23 on 2021-09-15 19:33

from django.db import migrations, models

def default_prior_entities(apps, schema_editor):
    entity_run_model = apps.get_model("django_cog", "EntityRun")
    for run in entity_run_model.objects.all():
        run.status = 'Completed'
        run.save()

class Migration(migrations.Migration):

    dependencies = [
        ('django_cog', '0016_auto_20210915_1042'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='cogerror',
            options={'ordering': ['-timestamp']},
        ),
        migrations.AddField(
            model_name='entityrun',
            name='status',
            field=models.CharField(choices=[('Queued', 'Queued'), ('Running', 'Running'), ('Failed', 'Failed'), ('Completed', 'Completed'), ('Canceled', 'Canceled')], default='Queued', max_length=10),
        ),
        migrations.RunPython(default_prior_entities)
    ]



import pytz
import time
from django.test import TestCase
from django.conf import settings
from .models import Pipeline, Stage, Task, Cog, PipelineRun, StageRun, TaskRun
from . import cog
from .apps import create_cog_records
from django_celery_beat.models import CrontabSchedule

class CogTestCase(TestCase):
    def setUp(self):
        @cog
        def test_function():
            return True
        
        create_cog_records()

    def test_cog_registration(self):
        """
        Test the function registration purely by
        checking if the name exists in our global
        "cog" object.
        """
        self.assertEquals((
            'test_function' in cog.all
        ), True)
    
    def test_cog_model_creation(self):
        """
        Test the function registration purely by
        checking if the name exists in our global
        "cog" object.
        """
        self.assertEquals(
            Cog.objects.filter(name='test_function').exists(),
            True
        )
